package at.ac.uibk.gitsearch.service;

import at.ac.uibk.gitsearch.domain.UserWatchList;
import at.ac.uibk.gitsearch.domain.WatchListEntry;
import at.ac.uibk.gitsearch.domain.enumeration.CheckFrequency;
import at.ac.uibk.gitsearch.repository.jpa.UserWatchListRepository;
import at.ac.uibk.gitsearch.repository.search.MetaDataRepository;
import at.ac.uibk.gitsearch.security.AuthoritiesConstants.AuthoritiesConstantEnum;
import at.ac.uibk.gitsearch.security.jwt.TokenProvider;
import at.ac.uibk.gitsearch.service.dto.UserWatchListDTO;
import at.ac.uibk.gitsearch.service.mapper.UserWatchListMapper;
import at.ac.uibk.gitsearch.web.rest.errors.BadRequestAlertException;
import java.time.Instant;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.concurrent.ConcurrentHashMap;
import java.util.stream.Collectors;
import org.codeability.sharing.plugins.api.search.SearchResultDTO;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.data.domain.Page;
import org.springframework.data.domain.PageRequest;
import org.springframework.data.domain.Pageable;
import org.springframework.data.domain.Sort;
import org.springframework.security.core.userdetails.User;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

/**
 * Service Implementation for managing {@link UserWatchList}.
 */
@Service
@Transactional
@SuppressWarnings("PMD.TooManyMethods")
public class UserWatchListService {

    private final Logger log = LoggerFactory.getLogger(UserWatchListService.class);

    private final UserWatchListRepository userWatchListRepository;

    private final UserWatchListMapper userWatchListMapper;

    private final TokenProvider tokenProvider;

    private final UserService userService;

    private final MetaDataRepository metaDataRepository;

    public UserWatchListService(
        UserWatchListRepository userWatchListRepository,
        UserWatchListMapper userWatchListMapper,
        TokenProvider tokenProvider,
        MetaDataRepository metaDataRepository,
        UserService userService
    ) {
        this.userWatchListRepository = userWatchListRepository;
        this.userWatchListMapper = userWatchListMapper;
        this.tokenProvider = tokenProvider;
        this.metaDataRepository = metaDataRepository;
        this.userService = userService;
    }

    /**
     * checks the access to this watchlist for the current user.
     * @param watchlistId the id of the watchlist
     * @param description some description of the request
     * @param mustExist if true, access fails if not existing
     * @throws BadRequestAlertException if access not allowed
     */
    @SuppressWarnings("PMD.CyclomaticComplexity")
    public Optional<UserWatchListDTO> checkAccessToWatchList(final Long watchlistId, final String description, boolean mustExist)
        throws IllegalAccessError {
        final Optional<User> currentPrincipal = tokenProvider.getCurrentPrincipal();
        if (currentPrincipal.isEmpty()) {
            log.warn("Cannot find a principal for watchlist {} for exercise {}", watchlistId, description);
            throw new IllegalAccessError("Cannot find a principal");
        }
        final Optional<UserWatchListDTO> watchListO = findOne(watchlistId);
        if (currentPrincipal.get().getAuthorities().contains(AuthoritiesConstantEnum.ADMIN.getGrantedAuthority())) {
            return watchListO; // ADMIN is always allowed
        }
        if (watchListO.isEmpty() && mustExist) {
            log.warn("Cannot find watchlist for : {} for exercise {}", watchlistId, description);
            throw new IllegalAccessError("Cannot find watchlist");
        }
        if (!mustExist) {
            return Optional.empty();
        }
        final boolean isAccessible = watchListO.get().getUserLogin().equals(currentPrincipal.get().getUsername());
        if (!isAccessible) {
            log.warn("watchlist {} does not belong to current user", watchListO.get().getName());
            throw new IllegalAccessError("watchlist does not belong to current User");
        }
        return watchListO; // ADMIN is always allowed
    }

    /**
     * Save a userWatchList.
     *
     * @param userWatchListDTO the entity to save.
     * @return the persisted entity.
     */
    public UserWatchListDTO save(UserWatchListDTO userWatchListDTO) {
        Optional<UserWatchList> previous = userWatchListDTO.getId() == null
            ? Optional.empty()
            : userWatchListRepository.findById(userWatchListDTO.getId());
        log.debug("Request to save UserWatchList : {}", userWatchListDTO);
        UserWatchList userWatchList = userWatchListMapper.toEntity(userWatchListDTO);
        if (previous.isPresent()) {
            userWatchList.setWatchListEntries(previous.get().getWatchListEntries());
        }
        userWatchList = userWatchListRepository.save(userWatchList);
        return userWatchListMapper.toDto(userWatchList);
    }

    /**
     * Get all the userWatchLists.
     *
     * @param pageable the pagination information.
     * @return the list of entities.
     */
    @Transactional(readOnly = true)
    public Page<UserWatchListDTO> findAll(Pageable pageable) {
        log.debug("Request to get all UserWatchLists");
        return userWatchListRepository.findAll(pageable).map(userWatchListMapper::toDto);
    }

    /**
     * Get all the userWatchLists.
     *
     * @param pageable the pagination information.
     * @return the list of entities.
     */
    @Transactional(readOnly = true)
    public List<UserWatchListDTO> findByUserIdIsCurrentUser() {
        log.debug("Request to get all UserWatchLists");
        return userWatchListRepository.findByUserIsCurrentUser().stream().map(userWatchListMapper::toDto).collect(Collectors.toList());
    }

    /**
     * Get one userWatchList by id.
     *
     * @param id the id of the entity.
     * @return the entity.
     */
    @Transactional(readOnly = true)
    public Optional<UserWatchListDTO> findOne(Long id) {
        log.debug("Request to get UserWatchList : {}", id);
        return userWatchListRepository.findById(id).map(userWatchListMapper::toDto);
    }

    /**
     * Delete the userWatchList by id.
     *
     * @param id the id of the entity.
     */
    public void delete(Long id) {
        log.debug("Request to delete UserWatchList : {}", id);
        userWatchListRepository.deleteById(id);
    }

    /**
     * Search for the userWatchList corresponding to the query.
     *
     * @param query the query of the search.
     * @param pageable the pagination information.
     * @return the list of entities.
     */
    @Transactional(readOnly = true)
    public Page<UserWatchListDTO> search(String query, Pageable pageable) {
        log.debug("Request to search for a page of UserWatchLists for query {}", query);
        final List<Sort.Order> cleanSortOrder = pageable
            .getSort()
            .get()
            .filter(order -> !"_score".equals(order.getProperty()))
            .collect(Collectors.toList());
        Pageable cleanedPageRequest = PageRequest.of(pageable.getPageNumber(), pageable.getPageSize(), Sort.by(cleanSortOrder));
        return userWatchListRepository.search(query, cleanedPageRequest).map(userWatchListMapper::toDto);
    }

    /**
     * Search for the userWatchList corresponding to the query.
     *
     * @param query the query of the search.
     * @param pageable the pagination information.
     * @return the list of entities.
     */
    @Transactional(readOnly = true)
    public Page<UserWatchListDTO> searchForCurrentUser(String query, Pageable pageable) {
        log.debug("Request to search for a page of CurrentUserWatchLists for query {}", query);
        final List<Sort.Order> cleanSortOrder = pageable
            .getSort()
            .get()
            .filter(order -> !"_score".equals(order.getProperty()))
            .collect(Collectors.toList());
        Pageable cleanedPageRequest = PageRequest.of(pageable.getPageNumber(), pageable.getPageSize(), Sort.by(cleanSortOrder));
        return userWatchListRepository.searchForCurrentUser(query, cleanedPageRequest).map(userWatchListMapper::toDto);
    }

    public List<UserWatchList> findByUser(at.ac.uibk.gitsearch.domain.User user) {
        return userWatchListRepository.findByUser(user);
    }

    /**
     * find all SearchResultDTO that are watched by some stored Watchlist, and
     * @param since
     * @param user
     * @return
     */
    public Map<SearchResultDTO, UserWatchList> findChangesSince(Instant since, at.ac.uibk.gitsearch.domain.User user) {
        Map<SearchResultDTO, UserWatchList> result = new ConcurrentHashMap<>();
        findByUser(user)
            .stream()
            .filter(wl -> wl.getCheckFrequency() != CheckFrequency.NEVER)
            .forEach(wl ->
                metaDataRepository
                    .getExercisesById(
                        wl.getWatchListEntries().stream().map(WatchListEntry::getExerciseId),
                        userService.convertToSecurityUser(Optional.of(user)),
                        0,
                        100,
                        since
                    )
                    .getSearchResult()
                    .forEach(sr -> {
                        UserWatchList previousWatchList = result.get(sr);
                        if (
                            previousWatchList == null || // none found or
                            previousWatchList.getCheckFrequency().ordinal() > wl.getCheckFrequency().ordinal()
                        ) { // earlier
                            result.put(sr, wl);
                        }
                    })
            );
        return result;
    }

    /**
     * not really useful.
     * @param checkFrequency
     * @param user
     * @return
     */
    public Map<SearchResultDTO, UserWatchList> findChangesSince(CheckFrequency checkFrequency, at.ac.uibk.gitsearch.domain.User user) {
        Map<SearchResultDTO, UserWatchList> changesSince = findChangesSince(checkFrequency.getChangedBefore(), user);
        return changesSince
            .entrySet()
            .stream()
            .filter(entry -> entry.getValue().getCheckFrequency().ordinal() <= checkFrequency.ordinal())
            .collect(Collectors.toMap(x -> x.getKey(), x -> x.getValue()));
    }
}
